#pragma once
#include <climits>
#include <thread>
#include <numeric>
#include <iterator>
#include <optional>
#include <vector>

typedef long double Num;
typedef std::vector<long double>::const_iterator NumIter;

//-----------------------------------------------------------------------------

void SumMapThread(NumIter begin, NumIter end, Num f(Num), long double &result)
{
    result = 0.0;
    for (NumIter it = begin; it != end; it++)
    {
        result += f(*it);
    }
}

/**
 * @brief Sums f(x) for x in [begin, end)
 * @param begin Start iterator
 * @param end End iterator
 * @param f Function to apply
 * @param result Reference to the result variable
 */
Num SumParallel(NumIter begin, NumIter end, Num f(Num), size_t num_threads)
{
    std::vector<std::thread> threads;
    std::vector<Num> results(num_threads);
    size_t step = std::distance(begin, end) / num_threads;
    for (size_t i = 0; i < num_threads; i++)
    {
        NumIter thread_begin = begin + i * step;
        NumIter thread_end = thread_begin + step;
        if (i == num_threads - 1)
        {
            thread_end = end;
        }
        threads.push_back(std::thread(SumMapThread, thread_begin, thread_end, f, std::ref(results[i])));
    }
    for (std::thread &t : threads)
    {
        t.join();
    }
    return std::accumulate(results.begin(), results.end(), 0.0);
}

//-----------------------------------------------------------------------------

/**
 * @brief Computes the mean of the numbers in [begin, end)
 * @param begin Start iterator
 * @param end End iterator
 * @param num_threads The number of threads to use
 * @return The mean in the range
 */
Num MeanParallel(NumIter begin, NumIter end, size_t num_threads)
{
    std::vector<std::thread> threads;
    std::vector<Num> results(num_threads);
    size_t step = std::distance(begin, end) / num_threads;

    for (size_t i = 0; i < num_threads; i++)
    {
        NumIter thread_begin = begin + i * step;
        NumIter thread_end = thread_begin + step;
        if (i == num_threads - 1)
        {
            thread_end = end;
        }

        threads.push_back(std::thread(SumMapThread, thread_begin, thread_end, [](Num x)
                                      { return x; }, std::ref(results[i])));
    }
    for (std::thread &t : threads)
    {
        t.join();
    }
    return std::accumulate(results.begin(), results.end(), 0.0) / std::distance(begin, end);
}

//-----------------------------------------------------------------------------

/**
 * @brief Computes the variance of the numbers in [begin, end)
 * @param begin Start iterator
 * @param end End iterator
 * @param num_threads The number of threads to use
 * @return The variance in the range
 */
Num VarianceParallel(NumIter begin, NumIter end, size_t num_threads)
{
    std::vector<std::thread> threads;
    std::vector<Num> results_x(num_threads, 0.0);
    std::vector<Num> results_x2(num_threads, 0.0);
    size_t step = std::distance(begin, end) / num_threads;

    for (size_t i = 0; i < num_threads; i++)
    {
        NumIter thread_begin = begin + i * step;
        NumIter thread_end = (i == num_threads - 1) ? end : thread_begin + step;

        // Compute sum of X
        threads.push_back(std::thread(SumMapThread, thread_begin, thread_end, [](Num x)
                                      { return x; }, std::ref(results_x[i])));

        // Compute sum of X^2
        threads.push_back(std::thread(SumMapThread, thread_begin, thread_end, [](Num x)
                                      { return x * x; }, std::ref(results_x2[i])));
    }
    for (std::thread &t : threads)
    {
        t.join();
    }

    Num mean_x = std::accumulate(results_x.begin(), results_x.end(), 0.0) / std::distance(begin, end);
    Num mean_x2 = std::accumulate(results_x2.begin(), results_x2.end(), 0.0) / std::distance(begin, end);

    return mean_x2 - (mean_x * mean_x);
}

//-----------------------------------------------------------------------------

/**
 * @brief Computs the occurences of the minimal value in [begin, end)
 * @param begin Start iterator
 * @param end End iterator
 * @param num_threads The number of threads to use
 * @return the number of occurences of the minimal value in [begin, end)
 */
int CountMinsParallel(std::vector<int>::const_iterator begin, std::vector<int>::const_iterator end, size_t num_threads)
{
    size_t length = std::distance(begin, end);
    if (length == 0)
        return 0;

    size_t step = length / num_threads;
    std::vector<int> min_values(num_threads, std::numeric_limits<int>::max());
    std::vector<int> min_counts(num_threads, 0);
    std::vector<std::thread> threads(num_threads);

    // Step 1: Find min values in each partition
    for (size_t i = 0; i < num_threads; ++i)
    {
        auto start = begin + i * step;
        auto stop = (i == num_threads - 1) ? end : start + step;

        threads[i] = std::thread([start, stop, &min_values, i]()
                                 { min_values[i] = *std::min_element(start, stop); });
    }

    for (size_t i = 0; i < num_threads; ++i)
    {
        threads[i].join();
    }

    // Step 2: Find the global minimum
    int global_min = *std::min_element(min_values.begin(), min_values.end());

    // Step 3: Reset threads and count occurrences of the global minimum
    for (size_t i = 0; i < num_threads; ++i)
    {
        auto start = begin + i * step;
        auto stop = (i == num_threads - 1) ? end : start + step;

        threads[i] = std::thread([start, stop, global_min, &min_counts, i]()
                                 { min_counts[i] = std::count(start, stop, global_min); });
    }

    for (size_t i = 0; i < num_threads; ++i)
    {
        threads[i].join();
    }

    return std::accumulate(min_counts.begin(), min_counts.end(), 0);
}

//-----------------------------------------------------------------------------

// Helper function to search for the target in a subrange
template <typename Iter, typename T>
bool FindInRange(Iter start, Iter stop, T target)
{
    for (auto it = start; it != stop; ++it)
    {
        if (*it == target)
        {
            return true;
        }
    }
    return false;
}

/**
 * @brief Finds target in [begin, end)
 * @param begin Start iterator
 * @param end End iterator
 * @param target The target to search for
 * @param num_threads The number of threads to use
 * @return The sum in the range
 */
template <typename Iter, typename T>
bool FindParallel(Iter begin, Iter end, T target, size_t num_threads)
{
    size_t length = std::distance(begin, end);
    if (length == 0)
        return false;

    size_t step = length / num_threads;
    std::vector<std::thread> threads(num_threads);
    std::vector<bool> results(num_threads, false);

    for (size_t i = 0; i < num_threads; ++i)
    {
        auto start = begin + i * step;
        auto stop = (i == num_threads - 1) ? end : start + step;

        threads[i] = std::thread([&results, i, start, stop, target]()
                                 { results[i] = FindInRange(start, stop, target); });
    }

    for (size_t i = 0; i < num_threads; ++i)
    {
        threads[i].join();
    }

    // Check if any of the threads found the target
    for (const auto &result : results)
    {
        if (result)
            return true;
    }

    return false;
}

//-----------------------------------------------------------------------------
