#include <algorithm>
#include <iostream>
#include <chrono>
#include <math.h>

//------------------------------------------------

double Sum(double* arr, size_t N) {
    double result = 0.;
    for (size_t i = 0; i < N; ++i) {
        result += arr[i];
    }
    return result;
}

//-------------------------------------------------

__global__
void warmup() {
    size_t index = blockIdx.x * blockDim.x + threadIdx.x;
}

//-------------------------------------------------

__global__
void sumKernel(double* d_arr, double* d_partialSums, size_t N) {
    extern __shared__ double sdata[];

    // Calculate thread ID
    size_t tid = threadIdx.x;
    size_t index = blockIdx.x * blockDim.x + threadIdx.x;

    // Load elements into shared memory
    sdata[tid] = (index < N) ? d_arr[index] : 0.0;
    __syncthreads();

    // Perform reduction in shared memory
    for (size_t s = blockDim.x / 2; s > 0; s >>= 1) {
        if (tid < s) {
            sdata[tid] += sdata[tid + s];
        }
        __syncthreads();
    }

    // Write the result of this block to the partial sums array
    if (tid == 0) {
        d_partialSums[blockIdx.x] = sdata[0];
    }
}

/**
 * @brief Computes the sum of the array
 * @param arr - the pointer to the beginning of an array
 * @param N - the length of the array
 */
 double SumGPU(double* arr, size_t N) {
    double* d_arr;
    double* d_partialSums;
    double result = 0.0;
    size_t size = N * sizeof(double);

    // Allocate device memory
    cudaMalloc((void**)&d_arr, size);

    // Copy data from host to device
    cudaMemcpy(d_arr, arr, size, cudaMemcpyHostToDevice);

    // Determine block and grid sizes
    int blockSize = 256;
    int numBlocks = (N + blockSize - 1) / blockSize;

    // Allocate memory for partial sums
    cudaMalloc((void**)&d_partialSums, numBlocks * sizeof(double));

    // Launch kernel to compute partial sums
    sumKernel<<<numBlocks, blockSize, blockSize * sizeof(double)>>>(d_arr, d_partialSums, N);

    // Copy partial sums back to host
    double* h_partialSums = (double*)malloc(numBlocks * sizeof(double));
    cudaMemcpy(h_partialSums, d_partialSums, numBlocks * sizeof(double), cudaMemcpyDeviceToHost);

    // Final summation on the host
    for (int i = 0; i < numBlocks; ++i) {
        result += h_partialSums[i];
    }

    // Free device and host memory
    cudaFree(d_arr);
    cudaFree(d_partialSums);
    free(h_partialSums);

    return result;
}

//---------------------------------------------------

int main(int argc, char* argv[]) {
    // setting the random seed to get the same result each time
    srand(42);

    // taking as input, which algo to run
    int alg_ind = std::stoi(argv[1]);

    // Generating data
    size_t N = 1 << 20;
    double* arr = (double*) malloc(N * sizeof(double));
    for (size_t i = 0; i < N; ++i) {
          arr[i] = static_cast <double> (rand()) / static_cast <double> (RAND_MAX);
    }

    warmup<<<1, 1>>>();

    double result = 0.;
    auto start = std::chrono::steady_clock::now();
    switch (alg_ind) {
        case 0: 
            result = Sum(arr, N);
            break;
        case 1:
            result = SumGPU(arr, N);
            break;
    }
    auto finish = std::chrono::steady_clock::now();
    auto elapsed = std::chrono::duration_cast<std::chrono::microseconds>(finish - start).count(); 
    std::cout << "Elapsed time: " << elapsed << std::endl;
    std::cout << "Total result: " << result << std::endl;
    
    delete[] arr;
    return 0;
}

//---------------------------------------------------
// when we run algo 0 with N = 1 << 26, the Elapsed time is 104832
// when we run algo 0 with N = 1 << 27, the Elapsed time is 98544
// but for 1 << 30 we get a time of 105060
// on the other hand, for 1 << 25, we have 98185, and for 1<<20, the value is similar

